# -*- coding: utf-8 -*-
#
# Copyright (C) 2017 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import flask
from sqlalchemy import exc as sa_exc
from sqlalchemy import sql
from sqlalchemy import orm

from dci.api.v1 import base, sso
from dci.auth import check_passwords_equal, decode_jwt
from dci import dci_config
from dci.common import exceptions as dci_exc
from dciauth.request import AuthRequest
from dciauth.signature import Signature
from dciauth.v2.headers import parse_headers
from dciauth.v2.signature import is_valid
from dci.db import models2
from dci.identity import Identity

from jwt import exceptions as jwt_exc


class BaseMechanism(object):
    def __init__(self, request):
        self.request = request
        self.identity = None

    def authenticate(self):
        """Authenticate the user, if the user fail to authenticate then the
        method must raise an exception with proper error message."""
        pass

    def get_user(self, model_constraint):
        try:
            return (
                flask.g.session.query(models2.User)
                .filter(models2.User.state == "active")
                .options(orm.selectinload("team"))
                .filter(model_constraint)
                .one()
            )
        except orm.exc.NoResultFound:
            return None

    def identity_from_user(self, user):
        user_info = {
            # UUID to str
            "id": str(user.id),
            "password": user.password,
            "name": user.name,
            "fullname": user.fullname,
            "timezone": user.timezone,
            "email": user.email,
            "sso_username": user.sso_username,
            "etag": user.etag,
            "is_user": True,
        }

        is_super_admin = False
        is_read_only_user = False
        is_epm_user = False
        user_teams = {}
        for user_team in user.team:
            if user_team.id == flask.g.team_admin_id:
                is_super_admin = True
            if user_team.id == flask.g.team_redhat_id:
                is_read_only_user = True
            if user_team.id == flask.g.team_epm_id:
                is_epm_user = True
            # TODO (gvincent): use user_team.serialize()
            user_teams[user_team.id] = {
                "id": user_team.id,
                "name": user_team.name,
                "has_pre_release_access": user_team.has_pre_release_access,
            }

        user_info["teams"] = user_teams
        user_info["is_super_admin"] = is_super_admin
        user_info["is_read_only_user"] = is_read_only_user
        user_info["is_epm_user"] = is_epm_user

        return Identity(user_info)

    def check_team_is_active(self, team_id):
        if self.identity.teams[team_id]["state"] != "active":
            name = self.identity.teams[team_id]["team_name"]
            raise dci_exc.DCIException("team %s not active" % name, status_code=412)


class BasicAuthMechanism(BaseMechanism):
    def authenticate(self):
        auth = self.request.authorization
        if not auth:
            raise dci_exc.DCIException("Authorization header missing", status_code=401)

        username = auth.username
        user = self.get_user(models2.User.name == username)
        if user is None:
            user = self.get_user(models2.User.email == username)
            if user is None:
                raise dci_exc.DCIException(
                    "User %s does not exists." % username, status_code=401
                )
        is_authenticated = check_passwords_equal(auth.password, user.password)
        if not is_authenticated:
            raise dci_exc.DCIException("Invalid user credentials", status_code=401)
        self.identity = self.identity_from_user(user)
        return True


class HmacMechanism(BaseMechanism):
    def authenticate(self):
        headers = self.request.headers
        auth_request = AuthRequest(
            method=self.request.method,
            endpoint=self.request.path,
            payload=self.request.get_json(silent=True),
            headers=headers,
            params=self.request.args.to_dict(flat=True),
        )
        hmac_signature = Signature(request=auth_request)
        self.identity = self.build_identity(auth_request.get_client_info())
        if self.identity is None:
            raise dci_exc.DCIException("identity does not exists.", status_code=401)
        secret = self.identity.api_secret
        if not hmac_signature.is_valid(secret):
            raise dci_exc.DCIException(
                "HmacMechanism failed: signature invalid",
                status_code=401,
            )
        if hmac_signature.is_expired():
            raise dci_exc.DCIException(
                "HmacMechanism failed: signature expired",
                status_code=401,
            )
        if len(self.identity.teams_ids) > 0:
            self.check_team_is_active(self.identity.teams_ids[0])
        return True

    def build_identity(self, client_info):
        allowed_types_model = {
            "remoteci": models2.Remoteci,
            "feeder": models2.Feeder,
        }
        client_type = client_info["client_type"]
        identity_model = allowed_types_model.get(client_type)
        if identity_model is None:
            return None
        identity = base.get_resource_orm(
            identity_model, client_info["client_id"], options=[orm.selectinload("team")]
        )
        return Identity(
            {
                "id": str(identity.id),
                "teams": {
                    identity.team.id: {
                        "team_name": identity.team.name,
                        "state": identity.team.state,
                        "has_pre_release_access": identity.team.has_pre_release_access,
                    }
                },
                "api_secret": str(identity.api_secret),
                "is_remoteci": client_type == "remoteci",
                "is_feeder": client_type == "feeder",
            }
        )


class Hmac2Mechanism(HmacMechanism):
    def authenticate(self):
        headers = parse_headers(self.request.headers)
        if not headers:
            raise dci_exc.DCIException(
                "Hmac2Mechanism failed: bad or incomplete headers.", status_code=400
            )
        self.identity = self.build_identity(headers)
        if self.identity is None:
            raise dci_exc.DCIException("identity does not exists.", status_code=401)
        valid, error_message = is_valid(
            {
                "method": self.request.method,
                "endpoint": self.request.path,
                "data": self.request.data,
                "params": self.request.args.to_dict(flat=True),
            },
            {"secret_key": self.identity.api_secret},
            headers,
        )
        if not valid:
            raise dci_exc.DCIException("Hmac2Mechanism failed: %s" % error_message)
        if len(self.identity.teams_ids) > 0:
            self.check_team_is_active(self.identity.teams_ids[0])
        return True


class OpenIDCAuth(BaseMechanism):
    def authenticate(self):
        auth_header = self.request.headers.get("Authorization").split(" ")
        if len(auth_header) != 2:
            return False
        bearer, token = auth_header

        conf = dci_config.CONFIG
        try:
            decoded_token = decode_jwt(
                token, conf["SSO_PUBLIC_KEY"], conf["SSO_CLIENT_ID"]
            )
        except (jwt_exc.DecodeError, ValueError):
            decoded_token = sso.decode_token_with_latest_public_key(token)
        except jwt_exc.ExpiredSignatureError:
            raise dci_exc.DCIException(
                "JWT token expired, please refresh.", status_code=401
            )

        team_id = None
        read_only_group = conf["SSO_READ_ONLY_GROUP"]
        if self._is_read_only_user(decoded_token, read_only_group):
            team_id = flask.g.team_redhat_id

        user_info = self._get_user_info(decoded_token)
        try:
            self.identity = self._get_or_update_or_create_user(user_info, team_id)
        except sa_exc.IntegrityError:
            raise dci_exc.DCICreationConflict("users", "username")
        return True

    @staticmethod
    def _is_read_only_user(token, read_only_group):
        # todo(gvincent): implement the solution with idp and verified email
        try:
            realm_access = token["realm_access"]
            return "roles" in realm_access and read_only_group in realm_access["roles"]
        except:
            return False

    @staticmethod
    def _get_user_info(token):
        # preferred_username is OIDC Standard
        username = token.get("preferred_username", token.get("username"))
        return {
            "name": username,
            "fullname": username,
            "sso_username": username,
            "email": token.get("email"),
            "timezone": "UTC",
        }

    def _get_or_update_or_create_user(self, user_info, team_id=None):
        sso_username = user_info["sso_username"]
        if not sso_username:
            raise Exception(
                "Red Hat login is required. Please contact a DCI administrator."
            )
        user = self._get_user_with_email_and_red_hat_login(user_info)
        if user:
            return self.identity_from_user(user)

        user = self._get_user_with_only_red_hat_login(user_info)
        if user:
            user_updated = self._update_user(user, user_info)
            return self.identity_from_user(user_updated)

        user_created = self._create_user(user_info, team_id)
        return self.identity_from_user(user_created)

    def _get_user_with_email_and_red_hat_login(self, user_info):
        sso_username = user_info["sso_username"]
        email = user_info["email"]
        existing_user_constraint = sql.and_(
            models2.User.sso_username == sso_username,
            models2.User.email == email,
        )
        return self.get_user(existing_user_constraint)

    def _get_user_with_only_red_hat_login(self, user_info):
        return self.get_user(models2.User.sso_username == user_info["sso_username"])

    def _create_user(self, user_info, team_id):
        try:
            user = models2.User(**user_info)
            flask.g.session.add(user)
            flask.g.session.commit()
            if team_id is not None:
                team = base.get_resource_orm(models2.Team, team_id)
                team.users.append(user)
                flask.g.session.add(team)
                flask.g.session.commit()
        except Exception:
            flask.g.session.rollback()
            raise dci_exc.DCIException(
                message="Cannot create user in Open ID Connect auth mechanism"
            )
        return self._get_user_with_email_and_red_hat_login(user_info)

    def _update_user(self, user, user_info):
        base.update_resource_orm(user, {"email": user_info["email"]})
        return self._get_user_with_email_and_red_hat_login(user_info)
